import torch
from torch import Tensor

from .shrinkage import trace


def is_positive_definite(A: Tensor) -> Tensor:
    """Efficiently check if `A` is p.d. by attempting Cholesky decomposition."""
    return torch.linalg.cholesky_ex(A).info.eq(0)


@torch.jit.script
def psd_sqrt(A: Tensor) -> Tensor:
    """Compute the unique p.s.d. square root of a positive semidefinite matrix."""
    L, U = torch.linalg.eigh(A)
    L = L[..., None, :].clamp_min(0.0)
    return U * L.sqrt() @ U.mH


def psd_sqrt_rsqrt(A: Tensor) -> tuple[Tensor, Tensor]:
    """Efficiently compute both the p.s.d. sqrt & pinv sqrt of p.s.d. matrix `A`."""
    L, U = torch.linalg.eigh(A)
    L = L[..., None, :].clamp_min(0.0)

    # Square root is easy
    sqrt = U * L.sqrt() @ U.mH

    # We actually compute the pseudo-inverse here for numerical stability.
    # Use the same heuristic as `torch.linalg.pinv` to determine the tolerance.
    thresh = L[..., None, -1] * A.shape[-1] * torch.finfo(A.dtype).eps
    rsqrt = U * torch.where(L > thresh, L.rsqrt(), 0.0) @ U.mH

    return sqrt, rsqrt


def ot_barycenter(
    Ks: Tensor, weights: Tensor | None = None, *, max_iter: int = 100
) -> Tensor:
    """Fixed-point iteration for the 2-Wasserstein barycenter of a set of Gaussians.

    Algorithm derived in "A fixed-point approach to barycenters in Wasserstein space"
    by Álvarez-Esteban et al. (2016) <https://arxiv.org/abs/1511.05355>.

    Args:
        Ks: `[n, d, d]` batch of covariance matrices, one for each centered Gaussian
            in the set whose barycenter we want to compute.
        weights: `[n]` batch of weights for each Gaussian.

    Returns:
        Covariance matrix of the barycenter.
    """
    n = len(Ks)
    assert n > 1, "Need at least two Gaussians to compute a barycenter"

    # Uniform weights by default
    if weights is None:
        weights = Ks.new_ones(n) / n
    else:
        assert len(weights) == n, "Need one weight per Gaussian"
        weights = weights / weights.sum()

    # Bookkeeping variables
    loss = torch.inf
    tol = torch.finfo(Ks.dtype).eps
    weights = weights.view(-1, 1, 1)  # Broadcastable to Ks

    # Initialize with arithmetic mean of covariance matrices
    mu = Ks.mul(weights).sum(dim=0)
    trace_avg = mu.trace()

    # Begin Álvarez-Esteban et al. fixed-point iteration
    for _ in range(max_iter):
        sqrt_mu, rsqrt_mu = psd_sqrt_rsqrt(mu)
        inner = psd_sqrt(sqrt_mu @ Ks @ sqrt_mu)

        # Equation 15 from Álvarez-Esteban et al. (2016)
        new_loss = mu.trace() + trace_avg - 2 * inner.mul(weights).sum(dim=0).trace()

        # Break if the loss is not decreasing
        if loss.real - new_loss.real < tol:
            break
        else:
            loss = new_loss

        # Equation 7 from Álvarez-Esteban et al. (2016)
        T = torch.sum(weights * rsqrt_mu @ inner @ rsqrt_mu, dim=0)
        mu = T @ mu @ T.mH

    return mu


def ot_distance(K1: Tensor, K2: Tensor) -> Tensor:
    """2-Wasserstein distance between N(0, K1) and N(0, K2)."""
    sqrt_K1 = psd_sqrt(K1)
    inner = psd_sqrt(sqrt_K1 @ K2 @ sqrt_K1)

    # Compute the 2-Wasserstein distance
    dist = torch.sqrt(trace(K1) + trace(K2) - 2 * trace(inner))
    return dist.squeeze(-1).squeeze(-1)


def ot_map(K1: Tensor, K2: Tensor) -> Tensor:
    """Optimal transport map from N(0, K1) to N(0, K2) in matrix form.

    Args:
        K1: Covariance matrix of the first Gaussian.
        K2: Covariance matrix of the second Gaussian.

    Returns:
        Unique p.s.d. matrix A such that N(0, A @ K1 @ A.T) = N(0, K2).
    """
    sqrt_K1, rsqrt_K1 = psd_sqrt_rsqrt(K1)
    return rsqrt_K1 @ psd_sqrt(sqrt_K1 @ K2 @ sqrt_K1) @ rsqrt_K1


def ot_midpoint(K1: Tensor, K2: Tensor, w1: float = 0.5, w2: float = 0.5) -> Tensor:
    """Covariance matrix of the 2-Wasserstein barycenter of N(0, K1) and N(0, K2).

    The barycenter of a set of distributions S is the unique distribution mu which
    minimizes the mean squared Wasserstein distance from each distribution in S to mu.

    Derived in "On Gaussian Wasserstein Barycenters" (Wessel Bruinsma & Gabriel Arpino)
    <https://gabrielarpino.github.io/files/wasserstein.pdf>.

    Args:
        K1: Covariance matrix of the first Gaussian.
        K2: Covariance matrix of the second Gaussian.
        w1: Weight of the first Gaussian.
        w2: Weight of the second Gaussian.

    Returns:
        Covariance matrix of the barycenter.
    """
    sqrt_K1, rsqrt_K1 = psd_sqrt_rsqrt(K1)
    product = sqrt_K1 @ psd_sqrt(sqrt_K1 @ K2 @ sqrt_K1) @ rsqrt_K1

    return w1 * w1 * K1 + w2 * w2 * K2 + w1 * w2 * (product + product.T)



from dataclasses import dataclass

import torch
from torch import Tensor

from .caching import cached_property, invalidates_cache
from .shrinkage import optimal_linear_shrinkage


@dataclass(frozen=True)
class OracleEraser:
    """Surgically erases a concept from a representation, given concept labels."""

    coef: Tensor
    mean_z: Tensor

    @classmethod
    def fit(cls, x: Tensor, z: Tensor, **kwargs) -> "OracleEraser":
        """Convenience method to fit an OracleEraser on data and return it."""
        return OracleFitter.fit(x, z, **kwargs).eraser

    def __call__(self, x: Tensor, z: Tensor) -> Tensor:
        """Replace `x` with the OLS residual given `z`."""
        # Ensure Z is at least 2D
        z = z.reshape(len(z), -1).type_as(x)
        expected_x = (z - self.mean_z) @ self.coef.T

        return x.sub(expected_x).type_as(x)


class OracleFitter:
    """Compute stats needed for surgically erasing a concept Z from a random vector X.

    Unlike `LeaceFitter`, the resulting erasure function requires oracle concept labels
    at inference time. In exchange, it achieves more surgical edits.
    """

    mean_x: Tensor
    """Running mean of X."""

    mean_z: Tensor
    """Running mean of Z."""

    sigma_xz_: Tensor
    """Unnormalized cross-covariance matrix X^T Z."""

    sigma_zz_: Tensor
    """Unnormalized covariance matrix Z^T Z."""

    n: Tensor
    """Number of X samples seen so far."""

    @classmethod
    def fit(cls, x: Tensor, z: Tensor, **kwargs) -> "OracleFitter":
        """Convenience method to fit a OracleFitter on data and return it."""
        n, d = x.shape
        _, k = z.reshape(n, -1).shape

        fitter = OracleFitter(d, k, device=x.device, dtype=x.dtype, **kwargs)
        return fitter.update(x, z)

    def __init__(
        self,
        x_dim: int,
        z_dim: int,
        *,
        device: str | torch.device | None = None,
        dtype: torch.dtype | None = None,
        shrinkage: bool = False,
        svd_tol: float = 0.01,
    ):
        """Initialize a `OracleFitter`.

        Args:
            x_dim: Dimensionality of the representation.
            z_dim: Dimensionality of the concept.
            method: Type of projection matrix to use.
            device: Device to put the statistics on.
            dtype: Data type to use for the statistics.
            shrinkage: Whether to use shrinkage to estimate the covariance matrix of X.
            svd_tol: Threshold for singular values of the covariance matrix of Z.
        """
        super().__init__()

        self.x_dim = x_dim
        self.z_dim = z_dim

        self.shrinkage = shrinkage

        assert svd_tol > 0.0, "`svd_tol` must be positive for numerical stability."
        self.svd_tol = svd_tol

        self.mean_x = torch.zeros(x_dim, device=device, dtype=dtype)
        self.mean_z = torch.zeros(z_dim, device=device, dtype=dtype)

        self.n = torch.tensor(0, device=device)
        self.sigma_xz_ = torch.zeros(x_dim, z_dim, device=device, dtype=dtype)
        self.sigma_zz_ = torch.zeros(z_dim, z_dim, device=device, dtype=dtype)

    @torch.no_grad()
    @invalidates_cache("eraser")
    def update(self, x: Tensor, z: Tensor) -> "OracleFitter":
        """Update the running statistics with a new batch of data."""
        d, c = self.sigma_xz_.shape
        x = x.reshape(-1, d).type_as(self.mean_x)
        n, d2 = x.shape

        assert d == d2, f"Unexpected number of features {d2}"
        self.n += n

        # Welford's online algorithm
        delta_x = x - self.mean_x
        self.mean_x += delta_x.sum(dim=0) / self.n

        z = z.reshape(n, -1).type_as(x)
        assert z.shape[-1] == c, f"Unexpected number of classes {z.shape[-1]}"

        delta_z = z - self.mean_z
        self.mean_z += delta_z.sum(dim=0) / self.n
        delta_z2 = z - self.mean_z

        self.sigma_xz_.addmm_(delta_x.mH, delta_z2)
        self.sigma_zz_.addmm_(delta_z.mH, delta_z2)

        return self

    @cached_property
    def eraser(self) -> OracleEraser:
        """Erasure function lazily computed given the current statistics."""
        return OracleEraser(
            self.sigma_xz @ torch.linalg.pinv(self.sigma_zz, atol=self.svd_tol),
            self.mean_z,
        )

    @property
    def sigma_zz(self) -> Tensor:
        """The covariance matrix of Z."""
        assert self.n > 1, "Call update() before accessing sigma_xx"
        assert (
            self.sigma_zz_ is not None
        ), "Covariance statistics are not being tracked for X"

        # Accumulated numerical error may cause this to be slightly non-symmetric
        S_hat = (self.sigma_zz_ + self.sigma_zz_.mH) / 2

        # Apply Random Matrix Theory-based shrinkage
        if self.shrinkage:
            return optimal_linear_shrinkage(S_hat / self.n, self.n)

        # Just apply Bessel's correction
        else:
            return S_hat / (self.n - 1)

    @property
    def sigma_xz(self) -> Tensor:
        """The cross-covariance matrix."""
        assert self.n > 1, "Call update() with labels before accessing sigma_xz"
        return self.sigma_xz_ / (self.n - 1)



import torch
from torch import Tensor


def optimal_linear_shrinkage(S_n: Tensor, n: int | Tensor) -> Tensor:
    """Optimal linear shrinkage for a sample covariance matrix or batch thereof.

    Given a sample covariance matrix `S_n` of shape (*, p, p) and a sample size `n`,
    this function computes the optimal shrinkage coefficients `alpha` and `beta`, then
    returns the covariance estimate `alpha * S_n + beta * Sigma0`, where ``Sigma0` is
    an isotropic covariance matrix with the same trace as `S_n`.

    The formula is distribution-free and asymptotically optimal in the Frobenius norm
    among all linear shrinkage estimators as the dimensionality `p` and sample size `n`
    jointly tend to infinity, with the ratio `p / n` converging to a finite positive
    constant `c`. The derivation is based on Random Matrix Theory and assumes that the
    underlying distribution has finite moments up to 4 + eps, for some eps > 0.

    See "On the Strong Convergence of the Optimal Linear Shrinkage Estimator for Large
    Dimensional Covariance Matrix" <https://arxiv.org/abs/1308.2608> for details.

    Args:
        S_n: Sample covariance matrices of shape (*, p, p).
        n: Sample size.
    """
    p = S_n.shape[-1]
    assert S_n.shape[-2:] == (p, p)

    # TODO: Make this configurable, try using diag(S_n) or something
    eye = torch.eye(p, dtype=S_n.dtype, device=S_n.device).expand_as(S_n)
    trace_S = trace(S_n)
    sigma0 = eye * trace_S / p

    sigma0_norm_sq = sigma0.norm(dim=(-2, -1), keepdim=True) ** 2
    S_norm_sq = S_n.norm(dim=(-2, -1), keepdim=True) ** 2

    prod_trace = trace(S_n @ sigma0)
    top = trace_S * trace_S.conj() * sigma0_norm_sq / n
    bottom = S_norm_sq * sigma0_norm_sq - prod_trace * prod_trace.conj()

    # Epsilon prevents dividing by zero for the zero matrix. In that case we end up
    # setting alpha = 0, beta = 1, but it doesn't matter since we're shrinking toward
    # tr(0)*I = 0, so it's a no-op.
    eps = torch.finfo(S_n.dtype).eps
    alpha = 1 - (top + eps) / (bottom + eps)
    beta = (1 - alpha) * (prod_trace + eps) / (sigma0_norm_sq + eps)

    return alpha * S_n + beta * sigma0


def trace(matrices: Tensor) -> Tensor:
    """Version of `torch.trace` that works for batches of matrices."""
    diag = torch.linalg.diagonal(matrices)
    return diag.sum(dim=-1, keepdim=True).unsqueeze(-1)



from dataclasses import dataclass
from typing import Callable, Iterator

import torch
from torch import LongTensor, Tensor


@dataclass(frozen=True)
class GroupedTensor:
    """A tensor split into groups along a given dimension.

    This class contains all the information needed to reconstruct the original tensor,
    or to take a list of tensors derived from the groups and coalesce them in such a
    way that the original order is restored.
    """

    dim: int
    """Dimension along which the tensor was split."""

    groups: list[Tensor]
    """List of tensors such that `groups[i]` contains all elements of `x` whose group
    label is `labels[i]`."""

    indices: LongTensor
    """Indices used to sort the original tensor."""

    labels: list[int]
    """Unique label for each element of `groups`."""

    def coalesce(self, groups: list[Tensor] | None = None) -> Tensor:
        """Fuse `groups or self.groups` back together, restoring the original order.

        This method is most useful when you want to group a tensor, perform an operation
        on each group, then combine the results back together.
        """
        if groups is None:
            groups = self.groups

        # First concatenate the groups back together
        fused = torch.cat(groups, dim=self.dim)

        # Invert the permutation to restore the original order
        return fused.index_select(self.dim, invert_indices(self.indices))

    def map(self, fn: Callable[[int, Tensor], Tensor]) -> "GroupedTensor":
        """Apply `fn` to each group & return a new `GroupedTensor` with the results."""
        results = [fn(label, group) for label, group in zip(self.labels, self.groups)]
        return GroupedTensor(self.dim, results, self.indices, self.labels)

    def __iter__(self) -> Iterator[tuple[int, Tensor]]:
        """Iterate over the groups and their labels."""
        for label, group in zip(self.labels, self.groups):
            yield label, group


def groupby(
    x: Tensor, key: Tensor, dim: int = 0, *, stable: bool = False
) -> GroupedTensor:
    """Efficiently split `x` into groups along `dim` according to `key`.

    This function is intended to mimic the behavior of `itertools.groupby`, but for
    PyTorch tensors. Under the hood, we sort `x` by `key` once, then return views
    onto the sorted tensor in order to minimize the number of memcpy and equality
    checking operations performed.

    By necessity this operation performs a host-device sync since we need to know
    the number of groups and their sizes in order to create a view for each.

    Args:
        x: Tensor to split into groups.
        key: Tensor of group labels.
        dim: Dimension along which to split `x`.
        stable: If `True`, use a stable sorting algorithm. This is slower but ensures
            that the order of elements within each group is preserved.

    Returns:
        A `GroupedTensor` containing the groups, sorting indices, and labels.
    """
    assert key.dtype == torch.int64, "`key` must be int64"
    assert key.ndim == 1, "`key` must be 1D"

    key, indices = key.sort(stable=stable)
    labels, counts = key.unique_consecutive(return_counts=True)

    # Sort `x` by `key` along `dim`
    x = x.index_select(dim, indices)
    groups = x.split(counts.tolist(), dim=dim)

    return GroupedTensor(dim, groups, indices, labels.tolist())


@torch.jit.script
def invert_indices(indices: Tensor) -> Tensor:
    """Efficiently invert the permutation represented by `indices`.

    Example:
        >>> indices = torch.tensor([2, 0, 1])
        >>> invert_indices(indices)
        tensor([1, 2, 0])
    """
    # Create an empty tensor to hold the reverse permutation
    reverse_indices = torch.empty_like(indices)

    # Scatter the indices to reverse the permutation
    arange = torch.arange(len(indices), device=indices.device)
    reverse_indices.scatter_(0, indices, arange)

    return reverse_indices



import pytest
import torch
from torch.distributions import Dirichlet

from concept_erasure.optimal_transport import (
    is_positive_definite,
    ot_barycenter,
    ot_distance,
    ot_map,
    ot_midpoint,
)


@pytest.mark.parametrize("d", [1, 2, 4, 8, 16, 32])
def test_gaussian_ot_map(d: int):
    torch.manual_seed(42)

    # Generate random positive definite matrices A and B
    A = torch.randn(d, d, dtype=torch.float64)
    A = A @ A.T / d

    B = torch.randn_like(A)
    B = B @ B.T / d

    C = torch.randn_like(A)
    C = C @ C.T / d

    # Compute the optimal transport map
    M = ot_map(A, B)

    # Analytically compute the covariance after transport
    torch.testing.assert_close(M @ A @ M, B)
    assert is_positive_definite(M)

    # Midpoint of two Gaussians, with different weights
    for w1 in [0.25, 0.5, 0.75]:
        w2 = 1 - w1

        mu = ot_midpoint(A, B, w1, w2)
        assert is_positive_definite(mu)

        # Check first order optimality condition
        mu.requires_grad_(True)
        loss = w1 * ot_distance(mu, A).square() + w2 * ot_distance(mu, B).square()
        loss.backward()

        torch.testing.assert_close(mu.grad, torch.zeros_like(mu), rtol=1e-6, atol=2e-6)

    # Barycenter of three Gaussians, with a few random weights
    weight_prior = Dirichlet(A.new_ones(3))
    for weights in weight_prior.sample(torch.Size([3])):
        Ks = torch.stack([A, B, C])
        mu = ot_barycenter(Ks, weights)
        assert is_positive_definite(mu)

        # Check first order optimality condition
        mu.requires_grad_(True)
        loss = ot_distance(mu, Ks).square().mul(weights).sum()
        loss.backward()
        torch.testing.assert_close(mu.grad, torch.zeros_like(mu), rtol=1e-6, atol=2e-6)



from contextlib import contextmanager
from functools import partial
from typing import TYPE_CHECKING

import torch
from torch import Tensor, nn

if (
    TYPE_CHECKING
):  # Don't import this unless we're type checking, since it's slow to import
    from transformers import PreTrainedModel

from ..utils import is_norm_layer


@contextmanager
def random_scrub(model: "PreTrainedModel", subspace_dim: int):
    """Add hooks to the model which erase a random subspace during `forward`."""
    d = model.config.hidden_size

    u = torch.empty(d, subspace_dim, device=model.device, dtype=model.dtype)
    nn.init.orthogonal_(u)

    def random_erase(_, __, x: Tensor, u: Tensor) -> Tensor:
        return x - u @ (u.T @ x)

    handles = [
        mod.register_forward_hook(partial(random_erase, u=u))
        for mod in model.modules()
        if is_norm_layer(mod)
    ]

    try:
        yield
    finally:
        # Make sure to remove the hooks even if an exception is raised
        for handle in handles:
            handle.remove()



import pytest
import torch

from concept_erasure import groupby


@pytest.mark.parametrize("stable", [True, False])
def test_groupby(stable: bool):
    n, k, d = 1000, 3, 10

    x = torch.randn(n, d)
    z = torch.randint(0, k, (n,))
    grouped = groupby(x, z, stable=stable)

    # Check that we can coalesce the groups back together
    torch.testing.assert_allclose(x, grouped.coalesce())
    torch.testing.assert_allclose(
        x + 1,
        grouped.map(lambda _, g: g + 1).coalesce(),
    )

    # We only expect these to be the same when stable=True
    if stable:
        for label in torch.unique(z):
            torch.testing.assert_allclose(
                x[z == label],
                grouped.groups[label],
            )



from .concept_scrubber import ConceptScrubber
from .groupby import GroupedTensor, groupby
from .leace import ErasureMethod, LeaceEraser, LeaceFitter
from .oracle import OracleEraser, OracleFitter
from .quadratic import QuadraticEditor, QuadraticEraser, QuadraticFitter
from .shrinkage import optimal_linear_shrinkage
from .utils import assert_type

__all__ = [
    "assert_type",
    "groupby",
    "optimal_linear_shrinkage",
    "ConceptScrubber",
    "ErasureMethod",
    "GroupedTensor",
    "LeaceEraser",
    "LeaceFitter",
    "OracleEraser",
    "OracleFitter",
    "QuadraticEditor",
    "QuadraticEraser",
    "QuadraticFitter",
]



from dataclasses import dataclass

import torch
from torch import Tensor

from .caching import cached_property, invalidates_cache
from .groupby import groupby
from .optimal_transport import ot_barycenter, ot_map, ot_midpoint
from .shrinkage import optimal_linear_shrinkage


@dataclass(frozen=True)
class QuadraticEraser:
    """Performs surgical quadratic concept erasure given oracle concept labels."""

    class_means: Tensor
    """`[k, d]` batch of class centroids."""

    global_mean: Tensor
    """`[d]` global centroid of the dataset."""

    ot_maps: Tensor
    """`[k, d, d]` batch of optimal transport matrices to the concept barycenter."""

    @classmethod
    def fit(cls, x: Tensor, z: Tensor, **kwargs) -> "QuadraticEraser":
        """Convenience method to fit a QuadraticEraser on data and return it."""
        return QuadraticFitter.fit(x, z, **kwargs).eraser

    def optimal_transport(self, z: int, x: Tensor) -> Tensor:
        """Transport `x` to the barycenter, assuming it was sampled from class `z`"""
        x_ = x.flatten(1)
        x_ = (x_ - self.class_means[z]) @ self.ot_maps[z].mH + self.global_mean
        return x_.view_as(x)

    def __call__(self, x: Tensor, z: Tensor) -> Tensor:
        """Apply erasure to `x` with oracle labels `z`."""
        # Efficiently group `x` by `z`, optimally transport each group, then coalesce
        return groupby(x, z).map(self.optimal_transport).coalesce()


@dataclass(frozen=True)
class QuadraticEditor:
    """Performs surgical quadratic concept editing."""

    class_means: Tensor
    """`[k, d]` batch of class centroids."""

    ot_maps: Tensor
    """`[k, k, d, d]` batch of pairwise optimal transport matrices between classes."""

    def transport(self, x: Tensor, source_z: int, target_z: int) -> Tensor:
        """Transport `x` from class `source_z` to class `target_z`"""
        T = self.ot_maps[source_z, target_z]
        return (x - self.class_means[source_z]) @ T.mH + self.class_means[target_z]

    def __call__(self, x: Tensor, source_z: Tensor, target_z: int) -> Tensor:
        """Transport `x` from classes `source_z` to class `target_z`."""
        return (
            groupby(x, source_z)
            .map(lambda src, x_grp: self.transport(x_grp, src, target_z))
            .coalesce()
        )


class QuadraticFitter:
    """Compute barycenter & optimal transport maps for a quadratic concept eraser."""

    mean_x: Tensor
    """Running mean of X."""

    mean_z: Tensor
    """Running mean of Z."""

    sigma_xx_: Tensor
    """Unnormalized class-conditional covariance matrices X^T X."""

    n: Tensor
    """Number of samples seen so far in each class."""

    @classmethod
    def fit(cls, x: Tensor, z: Tensor, **kwargs) -> "QuadraticFitter":
        """Convenience method to fit a OracleFitter on data and return it."""
        d = x.shape[-1]
        k = int(z.max()) + 1  # Number of classes

        fitter = QuadraticFitter(d, k, device=x.device, dtype=x.dtype, **kwargs)
        return fitter.update(x, z)

    def __init__(
        self,
        x_dim: int,
        num_classes: int,
        *,
        device: str | torch.device | None = None,
        dtype: torch.dtype | None = None,
        shrinkage: bool = True,
    ):
        """Initialize a `QuadraticFitter`.

        Args:
            x_dim: Dimensionality of the representation.
            num_classes: Number of distinct classes in the dataset.
            device: Device to put the statistics on.
            dtype: Data type to use for the statistics.
            store_covariance: Whether to store the precision matrices for each class in
                addition to the optimal transport maps in the eraser. This is necessary
                if you want to use the eraser to predict concept labels, or attempt to
                perform approximate erasure without oracle labels.
            shrinkage: Whether to use shrinkage to estimate the covariance matrix of X.
        """
        super().__init__()

        self.num_classes = num_classes
        self.shrinkage = shrinkage

        self.mean_x = torch.zeros(num_classes, x_dim, device=device, dtype=dtype)
        self.n = torch.zeros(num_classes, device=device)
        self.sigma_xx_ = torch.zeros(
            num_classes, x_dim, x_dim, device=device, dtype=dtype
        )

    def update(self, x: Tensor, z: Tensor) -> "QuadraticFitter":
        """Update the running statistics with a new batch of data."""
        x = x.flatten(0, -2).type_as(self.mean_x)

        for label, group in groupby(x, z, dim=0):
            self.update_single(group, label)

        return self

    @torch.no_grad()
    @invalidates_cache("eraser")
    def update_single(self, x: Tensor, z: int) -> "QuadraticFitter":
        """Update the running statistics with `x`, all sampled from class `z`."""
        x = x.flatten(0, -2).type_as(self.mean_x)

        self.n[z] += len(x)

        # Welford's online algorithm
        delta_x = x - self.mean_x[z]
        self.mean_x[z] += delta_x.sum(dim=0) / self.n[z]
        delta_x2 = x - self.mean_x[z]

        self.sigma_xx_[z].addmm_(delta_x.mH, delta_x2)

        return self

    def editor(self) -> QuadraticEditor:
        """Quadratic editor for the concept."""
        sigma = self.sigma_xx
        return QuadraticEditor(self.mean_x, ot_map(sigma[:, None], sigma))

    @cached_property
    def eraser(self) -> QuadraticEraser:
        """Erasure function lazily computed given the current statistics."""

        sigmas = self.sigma_xx

        # Compute Wasserstein barycenter of the classes
        if self.num_classes == 2:
            # Use closed form solution for the binary case
            class_prior = self.n / self.n.sum()
            center = ot_midpoint(sigmas[0], sigmas[1], *class_prior.tolist())
        else:
            # Use fixed point iteration for the general case
            center = ot_barycenter(self.sigma_xx, self.n)

        # Then compute the optimal transport maps from each class to the barycenter
        ot_maps = ot_map(sigmas, center)

        return QuadraticEraser(
            self.mean_x,
            self.mean_x.mean(dim=0),
            ot_maps,
        )

    @property
    def sigma_xx(self) -> Tensor:
        """Class-conditional covariance matrices of X."""
        assert torch.all(self.n > 1), "Some classes have < 2 samples"
        assert (
            self.sigma_xx_ is not None
        ), "Covariance statistics are not being tracked for X"

        # Accumulated numerical error may cause this to be slightly non-Hermitian
        S_hat = (self.sigma_xx_ + self.sigma_xx_.mH) / 2

        # Apply Random Matrix Theory-based shrinkage
        n = self.n.view(-1, 1, 1)
        if self.shrinkage:
            return optimal_linear_shrinkage(S_hat / n, n)

        # Just apply Bessel's correction
        else:
            return S_hat / (n - 1)



from typing import Any, Type, TypeVar, cast

from torch import nn

T = TypeVar("T")


def assert_type(typ: Type[T], obj: Any) -> T:
    """Assert that an object is of a given type at runtime and return it."""
    if not isinstance(obj, typ):
        raise TypeError(f"Expected {typ.__name__}, got {type(obj).__name__}")

    return cast(typ, obj)


def is_norm_layer(module: nn.Module) -> bool:
    """Return `True` if the module is a normalization layer."""
    cls_name = type(module).__name__
    return cls_name.endswith("LayerNorm") or cls_name.endswith("RMSNorm")


def mangle_module_path(name: str) -> str:
    """Mangle a module path to make it a valid key in a `nn.ModuleDict`."""
    # This is a weird edge case we probably don't need to support
    assert "-" not in name, "Module path cannot contain `-`"
    return name.replace(".", "-")
